/******************************************************************************/
/******************************************************************************/
/* Program: xa_mqs_db2                                                        */
/* Author:  Predrag Maksimovic - Max                                          */
/* Date:    08/18/2001                                                        */
/* Contact: pmaksimovic@2d3d.com                                              */
/******************************************************************************/
/* This examples shows how to use MQSeries and DB2 in a distributed           */
/* transaction. MQSeries queue manager serves as a TX transaction manager     */
/* and as an XA resource manager, while DB2 DBMS serves as an XA resource     */
/* manager only                                                               */
/* MQI API is used to access MQSeries, and CLI API is used to access DB2      */
/*                                                                            */
/* The program arguments are:                                                 */
/* #1 The name of the queue manager (the TX transaction manager)              */
/* #2 The name of the queue from which a message is received                  */
/* #3 The name of the DB2 data source                                         */
/* #4 The name of the table to which the message is inserted                  */
/* #5 The name of the actual column to which the message is inserted          */
/*                                                                            */
/* The program receives a message from the specified queue, and inserts it    */
/* into the specified database table - all within the same distributed        */
/* transaction                                                                */
/******************************************************************************/
/* Resource manager stanza:                                                   */
/* XAResourceManager:                                                         */
/* Name=data_source_name                                                      */
/* SwitchFile=switch_load_file_full_path                                      */
/* XAOpenString=database_alias[,username,password]                            */
/* ThreadOfControl=PROCESS                                                    */
/******************************************************************************/
/* Hint1: On WinNT, link this example program with mqm.lib and db2api.lib     */
/* Hint2: If doing SQL selects, turn the SQL_ATTR_CURSOR_HOLD statement       */
/*        attribute to SQL_CURSOR_HOLD_OFF (using SQLSetStmtAttr())           */
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/* Include general header files                                               */
/******************************************************************************/
#include <stdio.h>
/******************************************************************************/
/* Include MQSeries header files                                              */
/******************************************************************************/
#include <cmqc.h>
/******************************************************************************/
/* Include DB2 header files                                                   */
/******************************************************************************/
#include <sqlcli1.h>

/******************************************************************************/
/* Function: DisplayMQSeriesError                                             */
/* Displays MQSeries MQI error info                                           */
/******************************************************************************/
void DisplayMQSeriesError(MQLONG lReason)
{
    printf("MQI call failed with reason code %d\n", lReason);
}

/******************************************************************************/
/* Function: DisplayMQSeriesError                                             */
/* Displays DB2 CLI error info                                                */
/******************************************************************************/
void DisplayDB2Error(SQLSMALLINT iHandleType, SQLHANDLE iHandle)
{
      SQLCHAR     szState[5];                        /* SQL state code        */
      SQLCHAR     szMessage[SQL_MAX_MESSAGE_LENGTH]; /* error message text    */
      SQLSMALLINT iMessageLen;                       /* error mesasge len     */
      SQLINTEGER  iNativeError    = 0;               /* native error code     */
      SQLRETURN   rc              = 0;               /* return code           */

      rc = SQLGetDiagRec(iHandleType,                /* handle type           */
                         iHandle,                    /* handle                */
                         1,                          /* status record number  */
                         szState,                    /* SQL state code        */ 
                         &iNativeError,              /* native error code     */
                         szMessage,                  /* error message         */
                         SQL_MAX_MESSAGE_LENGTH,     /* message buffer length */
                         &iMessageLen                /* actual message length */
                         );
      if (rc == SQL_SUCCESS)
      {
	      printf ("CLI call failed:\n"
                  "Error: %s\n"
                  "State code %s\n"
                  "Native error code %d\n",
			      szMessage, szState, iNativeError);
      }
      else
      {
          printf("Error details could not be obtained\n");
      }

}

/******************************************************************************/
/* Function: DisplayMQSeriesError                                             */
/* Program entry point                                                        */
/******************************************************************************/
int main(int argc, char *argv[])                 
{
   /***************************************************************************/
   /* MQSeries variables                                                      */
   /***************************************************************************/
   MQOD     od              = {MQOD_DEFAULT};    /* object descriptor         */
   MQMD     md              = {MQMD_DEFAULT};    /* message descriptor        */
   MQGMO    gmo             = {MQGMO_DEFAULT};   /* get message options       */
   MQBO     bo              = {MQBO_DEFAULT};    /* begin options             */
   MQHCONN  hConn           = 0;                 /* handle to connection      */
   MQHOBJ   hObj            = 0;                 /* handle to object          */
   MQLONG   lOptions        = 0;                 /* options                   */
   MQLONG   lCompCode       = 0;                 /* completion code           */
   MQLONG   lReason         = 0;                 /* reason code               */
   MQLONG   lMsgBufLen      = 100;               /* message buffer length     */
   MQLONG   lMsgLen         = 0;                 /* message length received   */
   MQBYTE   byMsgBuf[100];                       /* message buffer (100 bytes)*/
   char*    szQueueManager  = NULL;              /* queue manager name        */
   char*    szQueue         = NULL;              /* queue name                */

   /***************************************************************************/
   /* DB2 CLI variables                                                       */
   /***************************************************************************/
   SQLHENV     henv            = 0;               /* environment handle       */
   SQLHDBC     hdbc            = 0;               /* connection handle        */
   SQLHSTMT    hstmt           = 0;               /* statement handle         */
   char*       szDSN           = NULL;            /* data source name         */
   char*       szTable         = NULL;            /* table name               */
   char*       szColumn        = NULL;            /* column name              */
   char        szSQLStatement[200];               /* SQL statement            */
   SQLRETURN   rc              = 0;               /* return code              */

   /***************************************************************************/
   /* general variables                                                       */
   /***************************************************************************/
   char     cSelection      = 'c';               /* end transaction selection */


   /***************************************************************************/
   /* Check input arguments                                                   */
   /***************************************************************************/
   if (argc != 6)
   {
	   printf("Usage: %s queue_manager_name queue_name "
		      "data_source_name table_name column_name\n",
		      argv[0]);
	   exit(-1);
   }
   else
   {
	   szQueueManager = argv[1];
	   szQueue        = argv[2];
	   szDSN          = argv[3];
	   szTable        = argv[4];
	   szColumn       = argv[5];
   }

   
   /***************************************************************************/
   /* Connect to queue manager                                                */
   /***************************************************************************/
   MQCONN(szQueueManager,
	      &hConn,
		  &lCompCode,
		  &lReason);

   if (lCompCode != MQCC_OK)
   {
       printf("MQCONN ended with reason code %li\n", lReason);
       exit(-1);
   }
   else
   {
	   printf("Connected to queue manager %s\n", szQueueManager);
   }

	
   /***************************************************************************/
   /* Start a new transaction                                                 */
   /***************************************************************************/
   MQBEGIN (hConn, &bo, &lCompCode, &lReason);

   if (lCompCode != MQCC_OK)
   {
      printf("MQBEGIN ended with reason code %li\n", lReason);
	  goto Exit_Program;
   }
   else
   {
      printf("Unit of work started\n");
   } 

   /***************************************************************************/
   /* Open the queue                                                          */
   /***************************************************************************/
   strncpy(od.ObjectName, szQueue, MQ_Q_NAME_LENGTH);
   lOptions = MQOO_INPUT_AS_Q_DEF + MQOO_FAIL_IF_QUIESCING;
   MQOPEN(hConn, &od, lOptions, &hObj, &lCompCode, &lReason);
   if (lCompCode != MQCC_OK)
   {
      printf("MQOPEN ended with reason code %li\n", lReason);
	  goto Exit_Program;
   }
   else
   {
      printf("Opened queue %s\n", szQueue);
   } 


   /***************************************************************************/
   /* Connect to DB2 step 1 - allocate environment handle                     */
   /***************************************************************************/
   rc = SQLAllocEnv(&henv);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLAllocEnv() failed\n");
       DisplayDB2Error(SQL_HANDLE_ENV, henv);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLAllocEnv succeeded. henv=%d\n", henv);
   }

   /***************************************************************************/
   /* Connect to DB2 step 2 - obtain the connection handle                    */
   /***************************************************************************/
   rc = SQLAllocConnect(henv, &hdbc);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLAllocConnect() failed\n");
       DisplayDB2Error(SQL_HANDLE_ENV, henv);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLAllocConnect succeeded. hdbc=%d\n", hdbc);
   }


   /***************************************************************************/
   /* Specify coordinated transaction usage for the new connection            */
   /***************************************************************************/

   rc = SQLSetConnectAttr(hdbc,
						  SQL_ATTR_CONNECTTYPE,
                          (SQLPOINTER) SQL_COORDINATED_TRANS,
                          0);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLSetConnectAttr(SQL_COORDINATED_TRANS) failed\n");
       DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLSetConnectAttr(SQL_COORDINATED_TRANS) succeeded\n");
   }

   
   /***************************************************************************/
   /* Specify two-phase commit protocol for the new connection                */
   /***************************************************************************/
   rc = SQLSetConnectAttr(hdbc,
						   SQL_ATTR_SYNC_POINT,
                           (SQLPOINTER) SQL_TWOPHASE,
                           0);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLSetConnectAttr(SQL_TWOPHASE) failed\n");
       DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLSetConnectAttr(SQL_TWOPHASE) succeeded\n");
   }

   /***************************************************************************/
   /* Turn the autocommit mode off for the new connection                     */
   /***************************************************************************/
   rc = SQLSetConnectAttr(hdbc,
						   SQL_ATTR_AUTOCOMMIT,
                           (SQLPOINTER) SQL_AUTOCOMMIT_OFF,
                           0);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLSetConnectAttr(SQL_AUTOCOMMIT_OFF) failed\n");
       DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLSetConnectAttr(SQL_AUTOCOMMIT_OFF) succeeded\n");
   }


   /***************************************************************************/
   /* Connect to DB2 step 3 - make the actual connection                      */
   /***************************************************************************/
   rc = SQLConnect(hdbc,
	               szDSN,
	               SQL_NTS,
	               NULL,
	               0,
	               NULL,
	               0);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLConnect failed for %s\n", szDSN);
       DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       goto Exit_Program;
   }
   else
   {
	   printf("Connected to data source %s\n", szDSN);
   }

   /***************************************************************************/
   /* Allocate a SQL statement handle                                         */
   /***************************************************************************/
   rc = SQLAllocStmt(hdbc, &hstmt);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLAllocStmt() failed\n");
       DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLAllocStmt succeeded. hstmt=%d\n", hstmt);
   }

   /***************************************************************************/
   /* Get a message                                                           */
   /***************************************************************************/
   gmo.Options = MQGMO_WAIT + MQGMO_SYNCPOINT;
   memset(byMsgBuf, 0, lMsgBufLen);
   MQGET(hConn, hObj, &md, &gmo,
	     lMsgBufLen, byMsgBuf, &lMsgBufLen,
         &lCompCode, &lReason);
   if (lCompCode != MQCC_OK)
   {
       printf("MQGET failed\n");
       DisplayMQSeriesError(lReason);
       goto Exit_Program;
   }
   else
   {
	   printf("Message retrieved from the queue %s\n", szQueue);
   }

   /***************************************************************************/
   /* Build the SQL statement                                                 */
   /***************************************************************************/
   sprintf(szSQLStatement, "INSERT INTO %s(%s) VALUES('%s')",
	       szTable, szColumn, (char*)byMsgBuf);
   printf("SQL statement to be executed: %s\n", szSQLStatement);

   /***************************************************************************/
   /* Execute the statement                                                   */
   /***************************************************************************/
   rc = SQLExecDirect(hstmt, szSQLStatement, SQL_NTS);
   if (rc != SQL_SUCCESS)
   {
      
       printf("SQLExecDirect() failed\n");
       DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       goto Exit_Program;
   }
   else
   {
	   printf("SQLExecDirect succeeded\n");
   }


   /***************************************************************************/
   /* End the transaction                                                     */
   /***************************************************************************/
   printf("Enter C or c to commit, any other key to rollback\n");
   cSelection = getchar();
   if (cSelection == 'c' || cSelection == 'C')
   {
       /***********************************************************************/
       /* Commit the transaction                                              */
       /***********************************************************************/
	   MQCMIT(hConn, &lCompCode, &lReason);
	   if (lCompCode != MQCC_OK)
	   {
           printf("MQCMIT failed\n");
           DisplayMQSeriesError(lReason);
           goto Exit_Program;
	   }
       else
       {
           printf("The transaction has been committed\n");
       }
   }
   else
   {
       /***********************************************************************/
       /* Rollback the transaction                                            */
       /***********************************************************************/
	   MQBACK(hConn, &lCompCode, &lReason);
	   if (lCompCode != MQCC_OK)
	   {
           printf("MQBACK failed\n");
           DisplayMQSeriesError(lReason);
           goto Exit_Program;
	   }
       else
       {
           printf("The transaction has been rolled back\n");
       }
   }



Exit_Program:

   /***************************************************************************/
   /* Close queue if opened                                                   */
   /***************************************************************************/
   if (hObj)
   {
      lOptions = 0;
      MQCLOSE(hConn, &hObj, lOptions, &lCompCode, &lReason);
     
	  if (lCompCode != MQCC_OK)
	  {
          printf("MQCLOSE failed\n");
          DisplayMQSeriesError(lReason);
	  }
      else
      {
          printf("Closed queue %s\n", szQueue);
      }
   }

   /***************************************************************************/
   /* Disconnect from queue manager if connected                              */
   /***************************************************************************/
   if (hConn)
   {
      MQDISC(&hConn, &lCompCode, &lReason);

	  if (lCompCode != MQCC_OK)
	  {
          printf("MQDISC failed\n");
          DisplayMQSeriesError(lReason);
	  }
      else
      {
          printf("Disconnected from queue manager %s\n", szQueueManager);
      }
   }

   /***************************************************************************/
   /* Try to disconnect from the database server                              */
   /***************************************************************************/
   if (hdbc)
   {
       rc = SQLDisconnect(hdbc);
       if (rc != SQL_SUCCESS)
       {
      
           printf("SQLDisconnect() failed\n");
           DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       }
       else
       {
	       printf("Disconnected from data source %s\n", szDSN);
       }
   }

   /***************************************************************************/
   /* Free the connection handle if allocated                                 */
   /***************************************************************************/
   if (hdbc)
   {
       rc = SQLFreeConnect(hdbc);
       if (rc != SQL_SUCCESS)
       {
      
           printf("SQLFreeConnect() failed\n");
           DisplayDB2Error(SQL_HANDLE_DBC, hdbc);
       }
       else
       {
	       printf("SQLFreeConnect succeeded\n");
           hdbc = 0;
       }
   }

   /***************************************************************************/
   /* Free the environment handle if allocated                                */
   /***************************************************************************/
   if (henv)
   {
       rc = SQLFreeEnv(henv);
       if (rc != SQL_SUCCESS)
       {
      
           printf("SQLFreeEnv() failed\n");
           DisplayDB2Error(SQL_HANDLE_ENV, henv);
       }
       else
       {
	       printf("SQLFreeEnv succeeded\n");
           henv = 0;
       }
   }
   
   /***************************************************************************/
   /* Exit program                                                            */
   /***************************************************************************/
   return 0;
}
